#ifndef SOLVER
#define SOLVER

#define DEBUG 0
#define CHECK 0 // check if clause_state is correct.

#ifdef RANK3
#define RANK 3
#define N_BASES 84
#endif

#ifdef RANK4
#define RANK 4
#define N_BASES 1820
#endif

#ifdef RANK5
#define RANK 5
#define N_BASES 53130
#endif

#define ELEMENTS (RANK * RANK)
#define N_BASES_SYM (N_BASES - RANK)        // exclude the given base vectors.
#define MAX_LITERALS (N_BASES + RANK + 100) // 100 is some additional threshold

#include <clingo.h>
#include <time.h>
#include "args.h"

extern my_args_t arguments;

#define TRUTH_VALUE(assignment, lit) lit > 0 ? assignment[lit] : (assignment[-lit] == truth_value_free ? truth_value_free : (assignment[-lit] == truth_value_false ? truth_value_true : truth_value_false))

typedef enum
{
    truth_value_free = 0,
    truth_value_true,
    truth_value_false,

} truth_value_t;

typedef int lit_t;

typedef struct
{
    int b1, b2;  // base literals which are the reason the clause was added
    int indexB1; // gives the index to the lit2clauses list from lit b1 pointing to the clause
    int indexB2; // similar to indexB1

    int e; // the element in b1 which should be replaced by an element of b2

    lit_t literals[RANK]; // one of the literals must be true if b1 and b2 are true
    int n;                // number of literals <= RANK

    int indices[RANK]; // indices to every list. -1 if not needed

    int nFalse;
    int nTrue;

    int p1; // pointer to first unassigned literal
    int p2; // pointer to last unassigned literal
    // TODO currently use indices to get position

    int satisfied;

    int indexFire; // -1 if not present in list. In ready2fire list if unit clause (i.e., if pos1 == pos2 and the current position not true)
} clause_t;        // All informations needed for solving. Information about the state

typedef struct
{
    clause_t *clauses;
    int n;       // number of clauses
    int maxN;    // maximal number of clauses
} clause_list_t; // All informations needed for solving

typedef struct
{
    int *elems;
    int n;    // current number of elems
    int maxN; // maximal possible number of elems
} int_list_t; // All informations needed for solving

/**
 * Contains several clauses, more precesily, for each base pair (where both literals are true),
 * it contains the "clauses" for the base exchange property.
 * 
 * For i=lit2clauses[l].elems[j] must satisfy that clauseList.elems[i].b1 = l or clauseList.elems[i].b2 = l
 * Also, for clauseList.elems[i].b1 there must be i \in lit2clauses[b1] (exactly ones)
 * 
 * Similarly, for update list and ready two fire list
 */
typedef struct
{
    clause_list_t clauseList;                  // array of clauses. Contains the actuall data
    int_list_t baseLits2clauses[MAX_LITERALS]; // for each literal clause indices which are added because the literal is true.
    int_list_t lit2clauses[MAX_LITERALS];      // for each literal a list of clause, which must be update (given by indices), when the literal is given a truthvalue.
    int_list_t ready2fire;
} clauses_state_t; // All informations needed for solving

// state information for individual solving threads
typedef struct
{
    truth_value_t *assignment;

    clauses_state_t sym_state;

    // statistical stuff
    long int calls;
    long int addedClauses;

    int previousDecisionLvl;
    int currentCutoffDecisionLvl;
    // for decision lvl cutoff

    clock_t propTime;
    clock_t maximilityCheckTime;
} state_t;

typedef enum
{
    present,
    notPresent,
    notKnown
} assumption_state_t;

typedef struct
{
    int threads;
    int turnOff;
    int staticClause;

    int decisionLvl;

    int printLearnedClauses;
    int frequency;
} inst_arguments_t;

typedef struct
{
    int threads;
    int relevantVertices;
    int turnOff;
    int staticClause;
} inst_init_t; // All informations to create instance such that we can start solving, but do not start yet

#define MAX_ASSUMPTIONS 1000

typedef struct
{
    double timeout;

    // stores result
    clingo_solve_handle_t *handle;
    clingo_model_t const *model;
    bool finished;
    bool sat;
    clock_t time;

    int fixedLiterals[N_BASES];
    int nFixed;

    int presentTuples[MAX_ASSUMPTIONS][RANK];
    int notPresentTuples[MAX_ASSUMPTIONS][RANK];
    int nPresentTuples, nNotPresentTuples;
} inst_solving_t; // All informations needed for solving

typedef struct propagator_t propagator_t;

typedef struct solver_instance
{
    clingo_control_t *control;
    inst_arguments_t arguments; // states for a single thread
    double timeout;             //The timeout for one solution step

    clock_t start; // Start of the program
    lit_t baseIndex2Lit[N_BASES];
    lit_t baseIndex2LitSolver[N_BASES];

    int solverLit2baseIndex[MAX_LITERALS];

    // next three only used for printing state and getting assumptions
    int isRelevantSolverLiteral[MAX_LITERALS];
    int solveLitMax;

    propagator_t *propData;

    int printDecisions;
    int decide; // True if we should select the next decision literal on our own.

    int minTransversals;
    int maxTransversals;
    int kPaving;
    int printClauses;
} rotta_inst_t;

// state information for the propagator
struct propagator_t
{
    rotta_inst_t *inst;
    state_t *states; // states for a single thread
    size_t states_size;
};

// TODO number of generated cubes and on which decision level.
// do something at a certain decision lvl i.e. call a function which has there own additional data, which can just see, if it should be further cubed or not. Gets the current assignment as assumption and the decision lvl

// inits all the static stuff which is the same for all solvers

void init(rotta_inst_t *inst);

// solve
void solve(rotta_inst_t *inst, inst_solving_t *solving);

void setThreads(rotta_inst_t *inst, int threads);

void init_static(void);

void setDecisionCutoffValue(rotta_inst_t *inst, int lvl);
void decisionLvlCutoff(clingo_propagate_control_t *control, propagator_t *data, int thread, bool *r);
void getCurrentAssignedBases(int *clause, int *n, int *presentLiterals, int *nPresent, rotta_inst_t *inst, truth_value_t *assignment);
void turnOffRestarts(rotta_inst_t *inst);

void addNewClauses(clauses_state_t *state, lit_t lit, truth_value_t *assignment, propagator_t *data);
void removeClauses(clauses_state_t *state, lit_t lit);
#endif